unit SFHelper;

{
Helper unit for the Soundfont manager. Coded completely by Ficedula   (ficedula@lycos.co.uk)
The SFMAN.PAS unit contains _only_ a direct translation of the API. In other words, it's all
up to you to deal with it. For small, fast apps - eg. a command line tool, or part of a game
 - that's OK. But for a non-critical Windows program, these classes provide an easier way to
 access SFManager without worrying about function tables, DLL handles, etc.
That said, for some functions - like finding out preset size - the SF Manager won't do it, so
you'll need to use this unit anyway.

Classes:
  TSFHelper: Enumerates soundfont devices and loads the DLL. Devices[] property returns
      TSFDevice objects. I don't know whether you can create more than one TSFHelper at a time
      or not - you wouldn't need to, anyway, since it returns system-wide information only.
  TSFDevice: The core of the unit. Returns information (capabilities, RAM size, loaded banks)
      and sets information (load presets, whole SF2 files, etc). Don't create these directly -
      let the TSFHelper class create them for you. Also, free the TSFDevice's before you free
      their 'parent' TSFHelper.
  TSFFile: Gets information on an actual SF2 file (no of presets, their names/sizes, which
      instruments they use, so on). This doesn't actually use the SF Manager, so you could
      use it even without the DLL's, though I don't know how much use it would be.

Initial version: 21/9/00
TSFFile added 23/9/00
}

interface

Uses Windows,SysUtils,Classes,SFMan,SFFile;

Type
  TSFBankContents = (sfPresets, sfBankPresets, sfNone);
{Meaning of these:
sfPresets =       This bank contains presets.
sfBankPresets =   This bank contains presets that are children of another bank.
sfNone =          Obvious}

  TSFBank = record
    Contents: TSFBankContents;
    Soundfont,Description:ShortString;
  end;
  //Soundfont is the actual filename. Description may be a filename, or a title taken from the bank.

  TSFPreset = record
    Bank,PSet:          Word;
//Bank,PSet: needed in case you want to load FROM one preset index TO another.
//If they're $FFFF (equiv. -1) then the loader function will ignore them
    Description,Source: Shortstring;
  end;
  //Source is generally the file the preset came from.

Const
  BLANK_PRESET: TSFPreset = (Description: ''; Source: '');
  BLANK_BANK:   TSFBank = (Contents: sfNone; Soundfont: ''; Description: '');

  SYSTEM_MEMORY_DEVICES: Array[0..0] of String = ('SB Live!');
{
What's the point of this, you might ask. Well, according to Creative, any device which has
the System Memory cap (sfSystemMemory below) DOES NOT return valid numbers to the function
QueryStaticSampleMemorySize. As the name implies, it should ask about STATIC memory - ie.
built onto the card, available ONLY to the soundcard. However, the SB Live! doesn't actually
conform - it has system memory but DOES accept the QueryStatic request. So the above array
lists devices that SHOULDN'T support QueryStatic but do. So far, SB Live only.
I suppose I can see the point - it retains compatibility with old AWE only programs. Maybe
their API should be designed better, then.
BTW, Creatives documentation is REALLY poor. Even when it's correct - like the SF2 file format -
it's still incomprehensible.
}

Type
  TSFDevCapsType = (sfSoftSynth, sfSystemMemory, sfHiddenDevice);
  TSFDevCaps = set of TSFDevCapsType;
  TSFMemory = record
    Max,Available: DWORD;
  end;

  TSFDevice = class
    private
      FDevNum:    Integer;
      FDevName:   String;
      IFace:      PSFMANL101API;
      FCaps:      TSFDevCaps;
      function GetBank(Index:Integer):TSFBank;
      function GetPreset(Bank,Pset:Integer):TSFPreset;
      procedure SetBank(Index:Integer;Val:TSFBank);
      procedure SetPreset(Bank,Pset:Integer;Val:TSFPreset);
    public
      property Banks[Index: Integer]: TSFBank read GetBank write SetBank;
      property Presets[Bank,Pset:Integer]: TSFPreset read GetPreset write SetPreset;
      function Available:Boolean; //is device available (ie. not playing music?)
      Constructor Create(Dev:Integer;SFManFace:PSFMANL101API);
      property Number:Integer read FDevNum; //device ID from the SF Manager
      property Caps: TSFDevCaps read FCaps;
      function Memory:TSFMemory;
      procedure ClearAll; //wipe memory clear
  end;

  TSFHelper = class
    private
      DLL:      THandle;
      Manager:  PSFMANAGER;
      IFace:    PSFMANL101API;
      FDevices: TStringList;
      FNumDevs: Integer;
      function GetDev(Index:Integer): TSFDevice;
    public
      Constructor Create;
      Destructor Destroy; Override;
      property NumDevices: Integer read FNumDevs;
      property DeviceNames: TStringList read FDevices;
      property Devices[Index:Integer]: TSFDevice read GetDev;
      function OK:Boolean;
  end;

  TFileInstrument = record
    Name:         String;
    NumSamples:   Word;
    Samples:      PWord;
  end;

  TFilePreset = record
    Name:             String;
    NumInstruments:   Word;
    Instruments:      PWord;
    Bank,Preset:      Byte;
  end;

  TSFFile = class
    private
      FPresets:     TList;
      FPresetSize:  TList;
      FInstruments: TList;
      FInstrSize:   TList;
      FSamples:     TList;
      FPresetNames: TStringList;
      FInstrNames:  TStringList;
      FSampleSize:  Integer;
      FActive:      Boolean;
      function GetPreset(Index:Integer): TFilePreset;
      function GetInstrument(Index:Integer): TFileInstrument;
      function GetNumPresets: Integer;
      function GetNumInstrs: Integer;
      procedure Clearup;
      function GetPresetIndex(Bank,PSet:Integer): Integer;
    public
      Constructor Create;
      Constructor CreateFromFile(FName:String);
      Destructor Destroy; Override;
      procedure LoadFromFile(FName:String);
      property Presets[Index:Integer]: TFilePreset read GetPreset;
      property Instruments[Index:Integer]: TFileInstrument read GetInstrument;
      property PresetIndex[Bank,PSet:Integer]: Integer read GetPresetIndex;
      property NumPresets: Integer read GetNumPresets;
      property NumInstruments: Integer read GetNumInstrs;
      property Active: Boolean read FActive; //True if a file was loaded OK
      function TotalSize(Presets:TList):Integer;
      function SizeofPresets(Presets:TList):TList;
      property SampleSize:Integer read FSampleSize;
  end;

function BankFromFile(FName:String): TSFBank;
function PresetFromFile(FName:String): TSFPreset;
{
These functions save you from declaring variables just so you can pass a filename into
the Banks[] / Presets[] arrays in TSFDevice.
Note: to clear a bank/preset, just set the array member to one of the BLANK_BANK or
BLANK_PRESET constants.
}

implementation

function TSFFile.GetPreset(Index:Integer): TFilePreset;
begin
If Index=-1 then begin
  ZeroMemory(@Result,Sizeof(Result));
  Result.Name := '??';
  Exit;
end;
Result.Name := FPresetNames[Index];
Result.NumInstruments := Integer(FPresetSize[Index]) and $FFFF;
Result.Bank := Integer(FPresetSize[Index]) shr 24;
Result.Preset := (Integer(FPresetSize[Index]) shr 16) and $FF;
Result.Instruments := FPresets[Index];
end;

function TSFFile.GetPresetIndex(Bank,PSet:Integer): Integer;
var
I:    Integer;
begin
Result := -1;
If Bank=SFMAN_PERCUSSIVE_BANK then Bank := 128;
For I := 0 to FPresetSize.Count-1 do
  If ( (Integer(FPresetSize[I]) shr 24)=Bank ) and
     ( ( (Integer(FPresetSize[I]) shr 16) and $FF)=PSet ) then begin
        Result := I; Exit;
  end;
end;

function TSFFile.GetInstrument(Index:Integer): TFileInstrument;
begin
Result.Name := FInstrNames[Index];
Result.NumSamples := Integer(FInstrSize[Index]);
Result.Samples := FInstruments[Index];
end;

function TSFFile.GetNumPresets: Integer;
begin
Result := FPresetNames.Count;
end;

function TSFFile.GetNumInstrs: Integer;
begin
Result := FInstrNames.Count;
end;

function TSFFile.SizeofPresets(Presets:TList):TList;
var
I,J,K,Sz:     Integer;
Samps,Instr:  TList;
PW:           PWord;
Tmp:          Pointer;
begin
Samps := TList.Create; Instr := TList.Create; Result := TList.Create;
For I := 0 to Presets.Count-1 do begin
  If Integer(Presets[I]) = -1 then begin
    Result.Add(Pointer(0));
    Continue;
  end;
  Instr.Clear; Samps.Clear;
  PW := FPresets[Integer(Presets[I])];
  For J := 1 to (Integer(FPresetSize[Integer(Presets[I])]) and $FFFF) do begin
    Tmp := Pointer(Word(PW^));
    Inc(PW);
    If Instr.IndexOf(Tmp)=-1 then Instr.Add(Tmp);
  end;

  For J := 0 to Instr.Count-1 do begin
    PW := FInstruments[Integer(Instr[J])];
    For K := 1 to Integer(FInstrSize[Integer(Instr[J])]) do begin
      Tmp := Pointer(Word(PW^));
      Inc(PW);
      If Samps.IndexOf(Tmp)=-1 then Samps.Add(Tmp);
    end;
  end;
  
  Sz := 0;
  Samps.Sort(ListSortCompare);
  For J := 0 to Samps.Count-1 do begin
    Inc(Sz,Integer(FSamples[2*Integer(Samps[J])]));
    K := Integer(FSamples[1+2*Integer(Samps[J])]);
    If ( (K shr 16) and 14) <> $0  then
      Inc(Sz,2*Integer(FSamples[2*(K and $FFFF)]));
  end;

  Result.Add(Pointer(Sz));

end;

Samps.Free; Instr.Free;
end;

function TSFFile.TotalSize(Presets:TList): Integer;
var
I,J:        Integer;
Samps,Instr:TList;
PW:         PWord;
Tmp:        Pointer;
begin
Samps := TList.Create; Instr := TList.Create;
For I := 0 to Presets.Count-1 do begin
  PW := FPresets[Integer(Presets[I])];
  For J := 1 to (Integer(FPresetSize[Integer(Presets[I])]) and $FFFF) do begin
    Tmp := Pointer(Word(PW^));
    Inc(PW);
    If Instr.IndexOf(Tmp)=-1 then Instr.Add(Tmp);
  end;
end;
For I := 0 to Instr.Count-1 do begin
  PW := FInstruments[Integer(Instr[I])];
  For J := 1 to Integer(FInstrSize[Integer(Instr[I])]) do begin
    Tmp := Pointer(Word(PW^));
    Inc(PW);
    If Samps.IndexOf(Tmp)=-1 then Samps.Add(Tmp);
  end;
end;
Result := 0;
Samps.Sort(ListSortCompare);
For I := 0 to Samps.Count-1 do begin
  Inc(Result,Integer(FSamples[2*Integer(Samps[I])]));
  J := Integer(FSamples[1+2*Integer(Samps[I])]);
  If ( (J shr 16) and 14) <> $0  then
    Inc(Result,2*Integer(FSamples[2*(J and $FFFF)]));
end;
Samps.Free; Instr.Free;
end;

Constructor TSFFile.CreateFromFile(FName:String);
begin
Create; LoadFromFile(FName);
end;

Procedure TSFFile.LoadFromFile(FName:String);
var
Mem:          TMemoryStream;
Fil:          TFileStream;
Num,St,Ps,
K,J,OthPos,
OldPos,Sz,I:  Integer;
Head1,Head2:  TsfPresetHeader;
Bag1,Bag2:    TsfPresetBag;
Gen:          TsfGenList;
Samps,Insts:  TList;
Inst,Inst2:   TsfInst;
IBag,IBag2:   TsfInstBag;
IGen:         TsfInstGenList;
Samp:         TSFSample;
PW:           PWord;
begin
{This procedure is big and badly written. It also isn't well documented, simply because
the file format is a nasty RIFF structure anyway. Look at the Creative docs if you
really want to figure it out. Just accept that this should load in the preset/instrument/
sample details from a SF2 file.
It SHOULD exit without any problems if the file is corrupt. Check
the Active property to see. But it's best if you use proper SF2 files
anyhow.}

FActive := False;

Mem := TMemoryStream.Create; Insts := TList.Create; Samps := TList.Create;
try
Clearup;
Fil := TFileStream.Create(FName,fmOpenRead or fmShareDenyNone);
Fil.Position := 0;
Fil.ReadBuffer(I,4);
If I<>FourCC('RIFF') then Exit;
Fil.ReadBuffer(Sz,4);
If Fil.Size-8<>Sz then Exit;
Fil.ReadBuffer(I,4);
If I<>FourCC('sfbk') then Exit;
Repeat
  Fil.ReadBuffer(I,4);
  If I<>FourCC('LIST') then Exit;
  Fil.ReadBuffer(Sz,4);
  Fil.ReadBuffer(I,4);
  If I=FourCC('sdta') then FSampleSize := Sz;
  If I<>FourCC('pdta') then Fil.Position := Fil.Position + Sz-4;
Until (I=FourCC('pdta')) or (Fil.Size-1 <= Fil.Position);
If Fil.Size-1 <= Fil.Position then Exit;
Mem.CopyFrom(Fil,Sz-4); Fil.Free;

  Mem.Position := 0;
  Mem.ReadBuffer(I,4);
  If I<>FourCC('phdr') then Exit;
  Mem.ReadBuffer(Sz,4);
  OldPos := Mem.Position;
  For K := 1 to (Sz div Sizeof(Head1))-1 do begin
    Mem.Position := OldPos;
    Mem.ReadBuffer(Head1,Sizeof(Head1));
    FPresetNames.Add(Head1.PresetName);
    OldPos := Mem.Position;
    Mem.ReadBuffer(Head2,Sizeof(Head2));
    Mem.Position := Sz+8;
    Mem.ReadBuffer(I,4);
    If I<>FourCC('pbag') then Exit;
    Mem.ReadBuffer(Num,4);
    OthPos := Mem.Position;
    For I := 0 to Head1.wPresetBagNdx do
      Mem.ReadBuffer(Bag1,Sizeof(Bag1));
    For I := Head1.wPresetBagNdx+1 to Head2.wPresetBagNdx do
      Mem.ReadBuffer(Bag2,Sizeof(Bag2));
    Mem.Position := OthPos+Num;
    Mem.ReadBuffer(I,4);
    If I=FourCC('pmod') then begin
      Mem.ReadBuffer(Num,4);
      Mem.Position := Mem.Position+Num;
    end else Exit;
    Mem.ReadBuffer(I,4);
    If I<>FourCC('pgen') then Exit;
    Mem.ReadBuffer(Num,4);
    OthPos := Mem.Position;
    Insts.Clear;
    For I := 0 to Bag1.wGenNdx-1 do Mem.ReadBuffer(Gen,Sizeof(Gen));
    For I := Bag1.wGenNdx to Bag2.wGenNdx-1 do begin
      Mem.ReadBuffer(Gen,Sizeof(Gen));
      If Gen.sfGenOper=GEN_INSTRUMENT then
        Insts.Add(Pointer(Gen.genAmount.wAmount));
    end;
    Insts.Sort(ListSortCompare);
    St := (Head1.wBank and $FF) shl 24;
    St := St or ((Head1.wPreset and $FF) shl 16);
    St := St or (Insts.Count and $FFFF);
    FPresetSize.Add(Pointer(St));
    GetMem(PW,Sizeof(Word) * Insts.Count);
    FPresets.Add(PW);
    For I := 0 to Insts.Count-1 do begin
      PW^ := Integer(Insts[I]);
      Inc(PW);
    end;
  end;

  Mem.Position := OthPos+Num;
  Mem.ReadBuffer(I,4);
  If I<>FourCC('inst') then Exit;
  Mem.ReadBuffer(Sz,4); OldPos := Mem.Position;
  Samps.Clear; OthPos := Mem.Position;
  For J := 0 to (Sz div Sizeof(Inst))-2 do begin
    Mem.Position := OldPos;
    Mem.ReadBuffer(Inst,Sizeof(Inst));
    OldPos := Mem.Position;
    FInstrNames.Add(Inst.InstName);

    Mem.ReadBuffer(Inst2,Sizeof(Inst2));
    Mem.Position := OthPos+Sz;
    Mem.ReadBuffer(I,4);
    If I<>FourCC('ibag') then Exit;
    Mem.ReadBuffer(Num,4);
    Ps := Mem.Position;
    If Num mod 4 <> 0 then Exit;
    For K := 0 to Inst.wInstBagNdx do Mem.ReadBuffer(IBag,Sizeof(IBag));
    For K := Inst.wInstBagNdx+1 to Inst2.wInstBagNdx do Mem.ReadBuffer(IBag2,Sizeof(IBag2));
    Mem.Position := Num+Ps;
    Mem.ReadBuffer(I,4);
    Mem.ReadBuffer(Ps,4);
    If I<>FourCC('imod') then Exit;
    Mem.Position := Mem.Position+Ps;
    St := Mem.Position;
    Mem.ReadBuffer(I,4);
    If I<>FourCC('igen') then Exit;
    Mem.ReadBuffer(I,4);
    Samps.Clear;
    For K := 0 to IBag.wInstGenNdx-1 do Mem.ReadBuffer(IGen,Sizeof(IGen));
    For K := IBag.wInstGenNdx to IBag2.wInstGenNdx-1 do begin
      Mem.ReadBuffer(IGen,Sizeof(IGen));
      If IGen.sfGenOper = GEN_SAMPLE then
        If Samps.IndexOf(Pointer(IGen.genAmount.wAmount))=-1 then
          Samps.Add(Pointer(IGen.genAmount.wAmount));
    end;
    Samps.Sort(ListSortCompare);
    FInstrSize.Add(Pointer(Samps.Count));
    GetMem(PW,Sizeof(Word) * Samps.Count);
    FInstruments.Add(PW);
    For I := 0 to Samps.Count-1 do begin
      PW^ := Word(Samps[I]);
      Inc(PW);
    end;
  end;
Mem.Position := St;
Mem.ReadBuffer(I,4);
Mem.ReadBuffer(Sz,4);
If I<>FourCC('igen') then Exit;
Mem.Position := Mem.Position + Sz;
Mem.ReadBuffer(I,4);
Mem.ReadBuffer(Sz,4);
If I<>FourCC('shdr') then Exit;
For I := 1 to (Sz div Sizeof(TsfSample))-1 do begin
  Mem.ReadBuffer(Samp,Sizeof(Samp));
  FSamples.Add(Pointer( (Samp.dwEnd - Samp.dwStart) * 2 ) );
  J := Samp.wSampleLink or (Samp.sfSampleType shl 16);
  FSamples.Add(Pointer(J));
{What's all this? As well as holding the sample size, we also have to hold the
link and type. If the sample is stereo, SampleType tells us that, and SampleLink
tells us which is its associated other channel sample. To play a stereo sample,
an instrument only needs to reference ONE channel - we are supposed to also include
the other channel ourselves.}
FActive := True;
end;
finally
Mem.Free; Insts.Free; Samps.Free;
end;
end;

Constructor TSFFile.Create;
begin
Inherited;
FPresets := TList.Create; FPresetSize := TList.Create;
FInstruments := TList.Create; FInstrSize := TList.Create;
FSamples := TList.Create; FPresetNames := TStringList.Create;
FInstrNames := TStringList.Create; FActive := False;
end;

procedure TSFFile.Clearup;
var
PW:   PWord;
I:    Integer;
begin
For I := 0 to FPresets.Count-1 do begin
  PW := FPresets[I];
  FreeMem(PW,Sizeof(Word) * Integer(FPresetSize[I]));
end;
For I := 0 to FInstruments.Count-1 do begin
  PW := FInstruments[I];
  FreeMem(PW,Sizeof(Word) * Integer(FInstrSize[I]));
end;
FPresets.Clear; FPresetSize.Clear; FInstruments.Clear;
FInstrSize.Clear; FSamples.Clear; FPresetNames.Clear;
FInstrNames.Clear;
end;

Destructor TSFFile.Destroy;
begin
Clearup;
FPresets.Free; FPresetSize.Free; FInstruments.Free;
FInstrSize.Free; FSamples.Free; FPresetNames.Free;
FInstrNames.Free;
Inherited;
end;

Constructor TSFDevice.Create(Dev:Integer;SFManFace:PSFMANL101API);
var
SFCaps:   CSFCAPSOBJECT;
begin
Inherited Create;
IFace := SFManFace; FDevNum := Dev;
IFace^.SF_GetDevCaps(Dev,@SFCaps);
FDevName := SFCaps.m_DevName;
FCaps := [];
If ( (SFCaps.m_DevCaps shr 31) and $1) = $1 then
  FCaps := FCaps + [sfSystemMemory];
If ( (SFCaps.m_DevCaps shr 30) and $1) = $1 then
  FCaps := FCaps + [sfSoftSynth];
If ( (SFCaps.m_DevCaps shr 29) and $1) = $1 then
  FCaps := FCaps + [sfHiddenDevice];
end;

procedure TSFDevice.ClearAll; //wipe memory clear
var
K,I,J:      Integer;
begin
{SB Live! cards are SOOOOO uncooperative in clearing memory. Firstly, clearing presets
can fail unless you do it in EXACT reverse order to the load - something that can't always
be arranged. Secondly, they sometimes report a loaded bank even after it's been cleared!
That's why it tries to clear everything multiple times - it's the only way to be anything
like sure you've got all the loaded objects. Besides, Clear operations have a very low
overhead, timewise.
If anyone finds a better way, PLEASE tell me. I don't like this method.}

For K:=1 to 3 do begin
For I := 0 to 256 do begin
  Banks[I] := BLANK_BANK;
  While Banks[I].Contents=sfPresets do
    Banks[I] := BLANK_BANK;
  If Banks[I].Contents=sfPresets then begin
    If Banks[I].Contents<>sfNone then
      For J := 0 to 127 do
        Presets[I,J] := BLANK_PRESET;
  end;
end;
I := SFMAN_PERCUSSIVE_BANK;
  If Banks[I].Contents=sfPresets then begin
      Banks[I] := BLANK_BANK;
      If Banks[I].Contents<>sfNone then
        For J := 0 to 127 do
          Presets[I,J] := BLANK_PRESET;
  end;
end;

end;

function TSFDevice.Memory:TSFMemory;
var
OK:   Boolean;
I:    Integer;
begin
If sfSystemMemory in FCaps then begin
  OK := False;
  For I := Low(SYSTEM_MEMORY_DEVICES) to High(SYSTEM_MEMORY_DEVICES) do
    If Pos(SYSTEM_MEMORY_DEVICES[I],FDevName)>0 then OK:=True;
  If Not OK then begin
    Result.Max := -1; Result.Available := -1; Exit;
  end;
end;
IFace^.SF_Open(FDevNum);
IFace^.SF_QueryStaticSampleMemorySize(FDevNum,@Result.Max,@Result.Available);
IFace^.SF_Close(FDevNum);
end;

procedure TSFDevice.SetPreset(Bank,Pset:Integer;Val:TSFPreset);
var
Buf:      CSFBUFFEROBJECT;
Loc,Loc2: CSFMIDILOCATION;
begin
IFace^.SF_Open(FDevNum);
try
Loc.m_BankIndex := Bank; Loc.m_PresetIndex := Pset;
If Bank=128 then Loc.m_BankIndex := SFMAN_PERCUSSIVE_BANK;
If (Val.Source='') then begin
  IFace^.SF_ClearLoadedPreset(FDevNum,@Loc);
  Exit;
end;
GetMem(Buf.m_Buffer,1024);
Buf.m_Size := 1024; Buf.m_SizeUsed := 0;
ZeroMemory(Buf.m_Buffer,1024);
Loc2 := Loc;
If Val.Bank <> $FFFF then Loc2.m_BankIndex := Val.Bank;
If Val.PSet <> $FFFF then Loc2.m_PresetIndex := Val.PSet;
StrPCopy(PChar(Buf.m_Buffer),Val.Source);
Buf.m_Flag := Word(SFFLAG_OPER_FILE);
IFace^.SF_LoadPreset(FDevNum,@Loc2,@Loc,@Buf);
finally
IFace^.SF_Close(FDevNum);
end;
end;

procedure TSFDevice.SetBank(Index:Integer;Val:TSFBank);
var
Buf:  CSFBUFFEROBJECT;
Loc:  CSFMIDILOCATION;
begin
IFace^.SF_Open(FDevNum);
try
If Index=0 then Index:=SFMAN_SYNTH_BANK;
Loc.m_BankIndex := Index; Loc.m_PresetIndex := $0;
If (Val.Soundfont='') then begin
  IFace^.SF_ClearLoadedBank(FDevNum,@Loc);
  Exit;
end;
GetMem(Buf.m_Buffer,1024);
Buf.m_Size := 1024; Buf.m_SizeUsed := 0;
ZeroMemory(Buf.m_Buffer,1024);
StrPCopy(PChar(Buf.m_Buffer),Val.Soundfont);
IFace^.SF_LoadBank(FDevNum,@Loc,@Buf);
finally
IFace^.SF_Close(FDevNum);
end;
end;

function TSFDevice.GetPreset(Bank,Pset:Integer):TSFPreset;
var
Buf:  CSFBUFFEROBJECT;
Loc:  CSFMIDILOCATION;
PC:   PChar;
begin
IFace^.SF_Open(FDevNum);
try
GetMem(Buf.m_Buffer,1024);
Buf.m_Size := 1024; Buf.m_SizeUsed := 0;
ZeroMemory(Buf.m_Buffer,1024);
Loc.m_BankIndex := Bank; Loc.m_PresetIndex := Pset;
IFace^.SF_GetLoadedPresetDescriptor(FDevNum,@Loc,@Buf);
PC := Buf.m_Buffer;
Result.Description := PC;
PC := PC + Length(Result.Description) + 1;
Result.Source := PC;
finally
IFace^.SF_Close(FDevNum);
end;
end;

function TSFDevice.GetBank(Index:Integer):TSFBank;
var
W:    Word;
Buf:  CSFBUFFEROBJECT;
Loc:  CSFMIDILOCATION;
begin
IFace^.SF_Open(FDevNum);
try
W := Index;
IFace^.SF_IsMidiBankUsed(FDevNum,@W);
If W=Index then Result.Contents := sfPresets
  Else if W=$FFFF then Result.Contents := sfNone
    Else Result.Contents := sfBankPresets;
GetMem(Buf.m_Buffer,1024);
Buf.m_Size := 1024; Buf.m_SizeUsed := 0;
ZeroMemory(Buf.m_Buffer,1024);
Loc.m_BankIndex := Index; Loc.m_PresetIndex := $0;
IFace^.SF_GetLoadedBankPathname(FDevNum,@Loc,@Buf);
Result.Soundfont := PChar(Buf.m_Buffer);
ZeroMemory(Buf.m_Buffer,1024);
IFace^.SF_GetLoadedBankDescriptor(FDevNum,@Loc,@Buf);
Result.Description := PChar(Buf.m_Buffer);
FreeMem(Buf.m_Buffer);
finally
IFace^.SF_Close(FDevNum);
end;
end;

function TSFDevice.Available:Boolean;
begin
Result := (IFace^.SF_IsDeviceFree(FDevNum) = SFERR_NOERR);
end;

Constructor TSFHelper.Create;
var
DW:       DWORD;
I:        Integer;
Caps:     CSFCAPSOBJECT;
begin
Inherited Create;
FDevices := TStringList.Create;

DLL := LoadLibrary(SF_MASTER_MANAGER_FILENAME);
Manager := GetProcAddress(DLL,SF_FUNCTION_TABLE_NAME);
If (DLL<>0) and (Manager<>nil) then begin
  DW := Manager^.SF_QueryInterface(ID_SFMANL101API,@IFace);
  If DW=SFERR_NOERR then begin
    IFace^.SF_GetNumDevs(@DW);
    FNumDevs := DW;
    For I := 1 to DW do begin
      Caps.m_SizeOf := Sizeof(Caps);
      IFace^.SF_GetDevCaps(I-1,@Caps);
      FDevices.Add(Caps.m_DevName);
    end;
  end;
end;
end;

Destructor TSFHelper.Destroy;
begin
FDevices.Free;
Inherited;
end;

function TSFHelper.GetDev(Index:Integer): TSFDevice;
begin
Result := TSFDevice.Create(Index,IFace);
end;

function TSFHelper.OK:Boolean;
begin
Result := (DLL<>0) and (IFace<>nil);
end;

function BankFromFile(FName:String): TSFBank;
begin
Result.Contents := sfBankPresets;
Result.Soundfont := FName;
Result.Description := '';
end;

function PresetFromFile(FName:String): TSFPreset;
begin
Result.Description := '';
Result.Source := FName;
Result.Bank := $FFFF;
Result.PSet := $FFFF;
end;

end.
